# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

import pytest
import numpy as np

from glue.core import ComponentID, Data, DataCollection

from .. import link_helpers as lh
from ..link_helpers import (LinkTwoWay, MultiLink,
                            LinkSame, LinkAligned)


R, D, L, B = (ComponentID('ra'), ComponentID('dec'),
              ComponentID('lon'), ComponentID('lat'))


def forwards(x):
    print('forwards input', x)
    return x * 3


def backwards(x):
    print('backwards input', x)
    return x / 3


def forwards_xy(x, y):
    print('forwards inputs', x, y)
    return x * 3, y * 5


def backwards_xy(x, y):
    print('backwards inputs', x, y)
    return x / 3, y / 5


def check_link(link, from_, to, using=None):
    assert link.get_from_ids() == from_
    assert link.get_to_id() == to
    if using:
        assert link.get_using() == using


def check_using(link, inp, out):
    np.testing.assert_array_almost_equal(link.get_using()(*inp), out)


def test_LinkTwoWay():
    result = list(LinkTwoWay(R, D, forwards, backwards))
    check_link(result[0], [R], D, forwards)
    check_link(result[1], [D], R, backwards)


def test_multilink_forwards():
    result = MultiLink([R, D], [L, B], forwards_xy, labels2=['x', 'y'])
    assert len(result) == 2
    check_link(result[0], [R, D], L)
    check_link(result[1], [R, D], B)
    check_using(result[0], (3, 4), 9)
    check_using(result[1], (3, 4), 20)


def test_multilink_backwards():
    result = MultiLink([R, D], [L, B], backwards=backwards_xy, labels1=['x', 'y'])
    assert len(result) == 2
    check_link(result[0], [L, B], R)
    check_link(result[1], [L, B], D)
    check_using(result[0], (9, 20), 3)
    check_using(result[1], (9, 20), 4)


def test_multilink_forwards_backwards():
    result = MultiLink([R, D], [L, B], forwards_xy, backwards_xy)
    assert len(result) == 4
    check_link(result[0], [R, D], L)
    check_link(result[1], [R, D], B)
    check_link(result[2], [L, B], R)
    check_link(result[3], [L, B], D)
    check_using(result[0], (3, 4), 9)
    check_using(result[1], (3, 4), 20)
    check_using(result[2], (9, 20), 3)
    check_using(result[3], (9, 20), 4)


def test_multilink_nofunc():
    with pytest.raises(TypeError) as exc:
        MultiLink([R, D], [L, B])
    assert exc.value.args[0] == "Must supply either forwards or backwards"


def test_linksame_string():
    """String inputs auto-converted to component IDs"""
    # ComponentLink does type checking to ensure conversion happens
    links = LinkSame('a', 'b')


def test_identity():
    assert lh.identity('3') == '3'


def test_toid():
    assert lh._toid('test').label == 'test'

    cid = ComponentID('test2')
    assert lh._toid(cid) is cid

    with pytest.raises(TypeError) as exc:
        lh._toid(None)


@pytest.mark.parametrize('ndim', [0, 1, 2])
def test_link_aligned(ndim):
    shp = tuple([2] * ndim)
    data1 = Data(test=np.random.random(shp))
    data2 = Data(test=np.random.random(shp))

    links = LinkAligned(data1, data2)
    dc = DataCollection([data1, data2])
    dc.add_link(links)

    for i in range(ndim):
        id0 = data1.pixel_component_ids[i]
        id1 = data2.pixel_component_ids[i]
        np.testing.assert_array_equal(data1[id0], data2[id1])



# Set up configuration variables

__all__ = ['custom_viewer', 'qglue', 'test']

import os

import sys

from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution('glue-core').version
except DistributionNotFound:
    __version__ = 'undefined'

from ._mpl_backend import MatplotlibBackendSetter
sys.meta_path.append(MatplotlibBackendSetter())

from glue.viewers.custom.helper import custom_viewer

# Load user's configuration file
from .config import load_configuration
env = load_configuration()

from .qglue import qglue

from .main import load_plugins  # noqa


def test(no_optional_skip=False):
    from pytest import main
    root = os.path.abspath(os.path.dirname(__file__))
    args = [root, '-x']
    if no_optional_skip:
        args.append('--no-optional-skip')
    return main(args=args)


from glue._settings_helpers import load_settings
load_settings()


# In PyQt 5.5+, PyQt overrides the default exception catching and fatally
# crashes the Qt application without printing out any details about the error.
# Below we revert the exception hook to the original Python one. Note that we
# can't just do sys.excepthook = sys.__excepthook__ otherwise PyQt will detect
# the default excepthook is in place and override it.


def handle_exception(exc_type, exc_value, exc_traceback):
    sys.__excepthook__(exc_type, exc_value, exc_traceback)


sys.excepthook = handle_exception



import os

from qtpy import QtWidgets, QtGui
from qtpy.QtCore import Qt, QRect

__all__ = ['QtSplashScreen']


class QtSplashScreen(QtWidgets.QWidget):

    def __init__(self, *args, **kwargs):

        super(QtSplashScreen, self).__init__(*args, **kwargs)

        self.resize(627, 310)
        self.setStyleSheet("background-color:white;")
        self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)

        self.center()

        self.progress = QtWidgets.QProgressBar()

        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.addStretch()
        self.layout.addWidget(self.progress)

        pth = os.path.join(os.path.dirname(__file__), '..', '..', 'logo.png')
        self.image = QtGui.QPixmap(pth)

    def set_progress(self, value):
        self.progress.setValue(value)
        QtWidgets.QApplication.processEvents()  # update progress bar

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        painter.drawPixmap(QRect(20, 20, 587, 229), self.image)

    def center(self):
        # Adapted from StackOverflow
        # https://stackoverflow.com/questions/20243637/pyqt4-center-window-on-active-screen
        frameGm = self.frameGeometry()
        screen = QtWidgets.QApplication.desktop().screenNumber(QtWidgets.QApplication.desktop().cursor().pos())
        centerPoint = QtWidgets.QApplication.desktop().screenGeometry(screen).center()
        frameGm.moveCenter(centerPoint)
        self.move(frameGm.topLeft())



import pytest

from glue.core import Data, DataCollection
from glue.core.link_helpers import identity
from glue.dialogs.autolinker.qt.autolinker import AutoLinkPreview
from glue.core.component_link import ComponentLink

# NOTE: the autolinker re-uses the main widget from the link editor so we don't
# need to test all the functionality - just things that are specific to the
# autolink preview.


class TestLinkEditor:

    def setup_method(self, method):

        self.data1 = Data(x=[1, 2, 3], y=[2, 3, 4], z=[6, 5, 4], label='data1')
        self.data2 = Data(a=[2, 3, 4], b=[4, 5, 4], c=[3, 4, 1], label='data2')
        self.data3 = Data(i=[5, 4, 3], j=[2, 2, 1], label='data3')

        self.data_collection = DataCollection([self.data1, self.data2, self.data3])

    @pytest.mark.parametrize('accept', [False, True])
    def test_basic(self, accept):

        # Set up an existing link
        link1 = ComponentLink([self.data1.id['x']], self.data2.id['c'])
        self.data_collection.add_link(link1)

        # Set up two suggested links

        def add(x, y):
            return x + y

        def double(x):
            return x * 2

        def halve(x):
            return x / 2

        link2 = ComponentLink([self.data2.id['a'], self.data2.id['b']], self.data3.id['j'], using=add)
        link3 = ComponentLink([self.data3.id['i']], self.data2.id['c'], using=double, inverse=halve)

        suggested_links = [link2, link3]

        dialog = AutoLinkPreview('test autolinker', self.data_collection, suggested_links)
        dialog.show()
        link_widget = dialog.link_widget

        link_widget.state.data1 = self.data1
        link_widget.state.data2 = self.data2

        assert link_widget.listsel_current_link.count() == 1

        link_widget.state.data1 = self.data3

        assert link_widget.listsel_current_link.count() == 2

        if accept:

            dialog.accept()

            links = self.data_collection.external_links

            assert len(links) == 3

            assert isinstance(links[0], ComponentLink)
            assert links[0].get_from_ids()[0] is self.data1.id['x']
            assert links[0].get_to_id() is self.data2.id['c']
            assert links[0].get_using() is identity

            assert isinstance(links[1], ComponentLink)
            assert links[1].get_from_ids()[0] is self.data2.id['a']
            assert links[1].get_from_ids()[1] is self.data2.id['b']
            assert links[1].get_to_id() is self.data3.id['j']
            assert links[1].get_using() is add

            assert isinstance(links[2], ComponentLink)
            assert links[2].get_from_ids()[0] is self.data3.id['i']
            assert links[2].get_to_id() is self.data2.id['c']
            assert links[2].get_using() is double
            assert links[2].get_inverse() is halve

        else:

            dialog.reject()

            links = self.data_collection.external_links

            assert len(links) == 1

            assert isinstance(links[0], ComponentLink)
            assert links[0].get_from_ids()[0] is self.data1.id['x']
            assert links[0].get_to_id() is self.data2.id['c']
            assert links[0].get_using() is identity



"""
This module implements polygon-square intersection using matplotlib. It is
twice as fast as Shapely for this specific case and avoids requiring another
dependency.
"""

import numpy as np

from matplotlib.path import Path
from matplotlib.transforms import Bbox


def square_polygon_intersection(xmin, xmax, ymin, ymax, x, y):
    poly = Path(list(zip(x, y)))
    box = Bbox([[xmin, ymin], [xmax, ymax]])
    try:
        clipped_poly = poly.clip_to_bbox(box)
    except ValueError:
        return [], []
    else:
        return clipped_poly.vertices[:, 0], clipped_poly.vertices[:, 1]


def polygon_area(x, y):
    x1 = x
    x2 = np.roll(x, -1)
    y1 = y
    y2 = np.roll(y, -1)
    return abs(0.5 * np.sum(x1 * y2 - x2 * y1))


def square_polygon_overlap_area(xmin, xmax, ymin, ymax, x, y):
    x, y = square_polygon_intersection(xmin, xmax, ymin, ymax, x, y)
    if len(x) == 0:
        return 0.
    else:
        return polygon_area(x, y)



import os
import sys
import traceback
from contextlib import contextmanager
from functools import wraps

__all__ = ['set_cursor', 'set_cursor_cm', 'messagebox_on_error',
           'die_on_error']


def set_cursor(shape):
    """Set the Qt cursor for the duration of a function call, and unset

    :param shape: Cursor shape to set.
    """
    def wrapper(func):
        @wraps(func)
        def result(*args, **kwargs):
            from glue.utils.qt import get_qapp  # Here to avoid circ import
            app = get_qapp()
            app.setOverrideCursor(shape)
            try:
                return func(*args, **kwargs)
            finally:
                app.restoreOverrideCursor()
        return result

    return wrapper


# TODO: Does this really belong in this module?
@contextmanager
def set_cursor_cm(shape):
    """Context manager equivalent for :func:`set_cursor`."""
    from glue.utils.qt import get_qapp
    app = get_qapp()
    app.setOverrideCursor(shape)
    try:
        yield
    finally:
        app.restoreOverrideCursor()


# TODO: We should be able to avoid defining these as classes and defining
# __call__ below, by using contextmanager.

class messagebox_on_error(object):

    def __init__(self, msg, sep='\n', exit=False):
        self.msg = msg
        self.sep = sep
        self.exit = exit

    def __call__(self, f):
        @wraps(f)
        def decorated(*args, **kwargs):
            # If we are in glue testing mode, just execute function
            if os.environ.get("GLUE_TESTING") == 'True':
                return f(*args, **kwargs)
            with self:
                return f(*args, **kwargs)
        return decorated

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, tb):

        if exc_type is None:
            return

        # Make sure application has been started
        from glue.utils.qt import get_qapp  # Here to avoid circular import
        get_qapp()

        m = "%s\n%s" % (self.msg, exc_val)
        detail = ''.join(traceback.format_exception(exc_type, exc_val, tb))
        if len(m) > 500:
            detail = "Full message:\n\n%s\n\n%s" % (m, detail)
            m = m[:500] + '...'

        from qtpy import QtWidgets
        from qtpy.QtCore import Qt

        qmb = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Critical, "Error", m)
        qmb.setDetailedText(detail)
        qmb.resize(400, qmb.size().height())
        qmb.setTextInteractionFlags(Qt.TextSelectableByMouse)
        qmb.exec_()

        if self.exit:
            sys.exit(1)

        # Just for cases where we are testing and patching sys.exit
        return True


class die_on_error(messagebox_on_error):

    def __init__(self, msg):
        super(die_on_error, self).__init__(msg, exit=True)



# Code adapted from:
#
# http://rowinggolfer.blogspot.de/2010/08/qtextedit-with-autocompletion-using.html
#
# and based on:
#
# http://qt-project.org/doc/qt-4.8/tools-customcompleter.html

from qtpy import QtGui, QtWidgets
from qtpy.QtCore import Qt


__all__ = ["CompletionTextEdit"]


class CompletionTextEdit(QtWidgets.QTextEdit):

    def __init__(self, parent=None):

        super(CompletionTextEdit, self).__init__(parent)

        self.setMinimumWidth(400)
        self.completer = None
        self.word_list = None

        self.moveCursor(QtGui.QTextCursor.End)

    def set_word_list(self, word_list):
        self.word_list = word_list
        self.set_completer(QtWidgets.QCompleter(word_list))

    def set_completer(self, completer):

        if self.completer:
            self.disconnect(self.completer, 0, self, 0)
        if not completer:
            return

        self.completer = completer

        self.completer.setWidget(self)
        self.completer.setCompletionMode(QtWidgets.QCompleter.PopupCompletion)
        self.completer.setCaseSensitivity(Qt.CaseInsensitive)
        self.completer.activated.connect(self.insert_completion)

    def insert_completion(self, completion):

        tc = self.textCursor()

        existing = self.text_under_cursor()

        completion = completion[len(existing):] + " "

        self.setTextCursor(tc)

        self.insertPlainText(completion)

        self.completer.setCompletionPrefix('')

    def text_under_cursor(self):
        tc = self.textCursor()
        text = self.toPlainText()
        pos1 = tc.position()
        if pos1 == 0 or text[pos1 - 1] == ' ':
            return ''
        else:
            sub = text[:pos1]
            if ' ' in sub:
                pos2 = sub.rindex(' ')
                return sub[pos2 + 1:]
            else:
                return sub

    # The following methods override methods in QTextEdit and should not be
    # renamed.

    def focusInEvent(self, event):
        if self.completer:
            self.completer.setWidget(self)
        QtWidgets.QTextEdit.focusInEvent(self, event)

    def keyPressEvent(self, event):

        if self.completer and self.completer.popup().isVisible():
            if event.key() in (
                    Qt.Key_Enter,
                    Qt.Key_Return,
                    Qt.Key_Escape,
                    Qt.Key_Tab,
                    Qt.Key_Backtab):
                event.ignore()
                return

        # Check if TAB has been pressed
        is_shortcut = event.key() == Qt.Key_Tab

        if not self.completer or not is_shortcut:
            QtWidgets.QTextEdit.keyPressEvent(self, event)
            return

        eow = "~!@#$%^&*()_+{}|:\"<>?,./;'[]\\-="

        completion_prefix = self.text_under_cursor()

        if not is_shortcut and (len(event.text()) == 0 or event.text()[-1:] in eow):
            self.completer.popup().hide()
            return

        self.completer.setCompletionPrefix(completion_prefix)
        popup = self.completer.popup()
        popup.setCurrentIndex(self.completer.completionModel().index(0, 0))

        cr = self.cursorRect()
        cr.setWidth(self.completer.popup().sizeHintForColumn(0) +
                    self.completer.popup().verticalScrollBar().sizeHint().width())
        self.completer.complete(cr)



from unittest.mock import patch

import os

from glue.config import SettingRegistry
from glue._settings_helpers import load_settings, save_settings


def test_roundtrip(tmpdir):

    settings = SettingRegistry()

    settings.add('STRING', 'green', str)
    settings.add('INT', 3, int)
    settings.add('FLOAT', 5.5, float)
    settings.add('LIST', [1, 2, 3], list)

    with patch('glue.config.settings', settings):
        with patch('glue.config.CFG_DIR', tmpdir.strpath):

            settings.STRING = 'blue'
            settings.INT = 4
            settings.FLOAT = 3.5
            settings.LIST = ['A', 'BB', 'CCC']

            settings.reset_defaults()

            assert settings.STRING == 'green'
            assert settings.INT == 3
            assert settings.FLOAT == 5.5
            assert settings.LIST == [1, 2, 3]

            settings.STRING = 'blue'
            settings.INT = 4
            settings.FLOAT = 3.5
            settings.LIST = ['A', 'BB', 'CCC']

            save_settings()

            assert os.path.exists(os.path.join(tmpdir.strpath, 'settings.cfg'))

            settings.reset_defaults()

            settings.STRING = 'red'
            settings.INT = 5

            # Loading settings will only change settings that have not been
            # changed from the defaults...
            load_settings()

            assert settings.STRING == 'red'
            assert settings.INT == 5
            assert settings.FLOAT == 3.5
            assert settings.LIST == ['A', 'BB', 'CCC']

            # ... unless the ``force=True`` option is passed
            load_settings(force=True)

            assert settings.STRING == 'blue'
            assert settings.INT == 4
            assert settings.FLOAT == 3.5
            assert settings.LIST == ['A', 'BB', 'CCC']



from ..state import MatplotlibLegendState
from glue.config import settings

from matplotlib.colors import to_rgba


class TestMatplotlibLegendState:
    def setup_method(self, method):
        self.state = MatplotlibLegendState()

    def test_draggable(self):
        self.state.location = 'draggable'
        assert self.state.draggable
        assert self.state.mpl_location == 'best'

    def test_no_draggable(self):
        self.state.location = 'lower left'
        assert not self.state.draggable
        assert self.state.mpl_location == 'lower left'

    def test_no_edge(self):
        self.state.show_edge = False
        assert self.state.edge_color is None

    def test_default_color(self):
        assert self.state.frame_color == settings.BACKGROUND_COLOR
        assert self.state.edge_color == to_rgba(settings.FOREGROUND_COLOR, self.state.alpha)



from glue.config import layer_artist_maker
from glue.core.application_base import Application
from glue.core import Data
from glue.viewers.common.viewer import Viewer
from glue.viewers.common.layer_artist import LayerArtist


def test_custom_layer_artist_maker():

    # This is a test of the infrastructure that allows users/developers to
    # define custom layer artist makers - i.e. functions that given a dataset
    # and a viewer can decide whether to create a custom layer artist rather
    # than the default types available for a given viewer.

    class CustomLayerArtist(LayerArtist):
        pass

    class CustomApplication(Application):
        def add_widget(self, *args, **kwargs):
            pass

    @layer_artist_maker('custom_maker')
    def custom_maker(viewer, data):
        if hasattr(data, 'custom'):
            return CustomLayerArtist(viewer.state, layer=data)

    app = CustomApplication()

    data1 = Data(x=[1, 2, 3], label='test1')
    data2 = Data(x=[1, 2, 3], label='test2')
    data2.custom = True

    app.data_collection.append(data1)
    app.data_collection.append(data2)

    viewer = app.new_data_viewer(Viewer)

    assert len(viewer.layers) == 0

    # NOTE: Check exact type, not using isinstance
    viewer.add_data(data1)
    assert len(viewer.layers) == 1
    assert type(viewer.layers[0]) is LayerArtist

    viewer.add_data(data2)
    assert len(viewer.layers) == 2
    assert type(viewer.layers[1]) is CustomLayerArtist



# A plugin to enable link helpers for Astronomical coordinate conversions.

from astropy import units as u
from astropy.coordinates import (ICRS, FK5, FK4, Galactic, Galactocentric,
                                 galactocentric_frame_defaults)

from glue.core.link_helpers import BaseMultiLink
from glue.config import link_helper


__all__ = ["BaseCelestialMultiLink", "Galactic_to_FK5", "FK4_to_FK5",
           "ICRS_to_FK5", "Galactic_to_FK4", "ICRS_to_FK4",
           "ICRS_to_Galactic"]


class BaseCelestialMultiLink(BaseMultiLink):

    display = None
    frame_in = None
    frame_out = None

    def forwards(self, in_lon, in_lat):
        cin = self.frame_in(in_lon * u.deg, in_lat * u.deg)
        cout = cin.transform_to(self.frame_out)
        return cout.spherical.lon.degree, cout.spherical.lat.degree

    def backwards(self, out_lon, out_lat):
        cout = self.frame_out(out_lon * u.deg, out_lat * u.deg)
        cin = cout.transform_to(self.frame_in)
        return cin.spherical.lon.degree, cin.spherical.lat.degree

    # Backward-compatibility with glue-core <0.15
    forward = forwards
    backward = backwards


@link_helper(category='Astronomy')
class Galactic_to_FK5(BaseCelestialMultiLink):
    description = 'Link Galactic and FK5 (J2000) Equatorial coordinates'
    labels1 = 'l', 'b'
    labels2 = 'ra (fk5)', 'dec (fk5)'
    display = "Galactic <-> FK5 (J2000)"
    frame_in = Galactic
    frame_out = FK5


@link_helper(category='Astronomy')
class FK4_to_FK5(BaseCelestialMultiLink):
    description = 'Link FK4 (B1950) and FK5 (J2000) Equatorial coordinates'
    labels1 = 'ra (fk4)', 'dec (fk4)'
    labels2 = 'ra (fk5)', 'dec (fk5)'
    display = "FK4 (B1950) <-> FK5 (J2000)"
    frame_in = FK4
    frame_out = FK5


@link_helper(category='Astronomy')
class ICRS_to_FK5(BaseCelestialMultiLink):
    description = 'Link ICRS and FK5 (J2000) Equatorial coordinates'
    labels1 = 'ra (fk4)', 'dec (fk4)'
    labels2 = 'ra (icrs)', 'dec (icrs)'
    display = "ICRS <-> FK5 (J2000)"
    frame_in = ICRS
    frame_out = FK5


@link_helper(category='Astronomy')
class Galactic_to_FK4(BaseCelestialMultiLink):
    description = 'Link Galactic and FK4 (B1950) Equatorial coordinates'
    labels1 = 'l', 'b'
    labels2 = 'ra (fk4)', 'dec (fk4)'
    display = "Galactic <-> FK4 (B1950)"
    frame_in = Galactic
    frame_out = FK4


@link_helper(category='Astronomy')
class ICRS_to_FK4(BaseCelestialMultiLink):
    description = 'Link ICRS and FK4 (B1950) Equatorial coordinates'
    labels1 = 'ra (icrs)', 'dec (icrs)'
    labels2 = 'ra (fk4)', 'dec (fk4)'
    display = "ICRS <-> FK4 (B1950)"
    frame_in = ICRS
    frame_out = FK4


@link_helper(category='Astronomy')
class ICRS_to_Galactic(BaseCelestialMultiLink):
    description = 'Link ICRS and Galactic coordinates'
    labels1 = 'ra (icrs)', 'dec (icrs)'
    labels2 = 'l', 'b'
    display = "ICRS <-> Galactic"
    frame_in = ICRS
    frame_out = Galactic


@link_helper(category='Astronomy')
class GalactocentricToGalactic(BaseMultiLink):

    description = 'Link 3D Galactocentric and Galactic coordinates'
    labels1 = 'x (kpc)', 'y (kpc)', 'z (kpc)'
    labels2 = 'l (deg)', 'b (deg)', 'distance (kpc)'
    display = "3D Galactocentric <-> Galactic"

    def forwards(self, x_kpc, y_kpc, z_kpc):
        with galactocentric_frame_defaults.set('pre-v4.0'):
            gal = Galactocentric(x=x_kpc * u.kpc, y=y_kpc * u.kpc, z=z_kpc * u.kpc).transform_to(Galactic())
        return gal.l.degree, gal.b.degree, gal.distance.to(u.kpc).value

    def backwards(self, l_deg, b_deg, d_kpc):
        with galactocentric_frame_defaults.set('pre-v4.0'):
            gal = Galactic(l=l_deg * u.deg, b=b_deg * u.deg, distance=d_kpc * u.kpc).transform_to(Galactocentric())
        return gal.x.to(u.kpc).value, gal.y.to(u.kpc).value, gal.z.to(u.kpc).value



from collections import OrderedDict

import numpy as np

from glue.core import Subset
from echo import delay_callback
from glue.viewers.matplotlib.state import (MatplotlibDataViewerState,
                                           MatplotlibLayerState,
                                           DeferredDrawCallbackProperty as DDCProperty,
                                           DeferredDrawSelectionCallbackProperty as DDSCProperty)
from glue.core.data_combo_helper import ManualDataComboHelper, ComponentIDComboHelper
from glue.utils import defer_draw, nanmin, nanmax
from glue.core.link_manager import is_convertible_to_single_pixel_cid
from glue.core.exceptions import IncompatibleDataException

__all__ = ['ProfileViewerState', 'ProfileLayerState']


FUNCTIONS = OrderedDict([('maximum', 'Maximum'),
                         ('minimum', 'Minimum'),
                         ('mean', 'Mean'),
                         ('median', 'Median'),
                         ('sum', 'Sum')])


class ProfileViewerState(MatplotlibDataViewerState):
    """
    A state class that includes all the attributes for a Profile viewer.
    """

    x_att_pixel = DDCProperty(docstring='The component ID giving the pixel component '
                                  'shown on the x axis')

    x_att = DDSCProperty(docstring='The component ID giving the pixel or world component '
                                   'shown on the x axis')

    reference_data = DDSCProperty(docstring='The dataset that is used to define the '
                                            'available pixel/world components, and '
                                            'which defines the coordinate frame in '
                                            'which the images are shown')

    function = DDSCProperty(docstring='The function to use for collapsing data')

    normalize = DDCProperty(False, docstring='Whether to normalize all profiles '
                                             'to the [0:1] range')

    # TODO: add function to use

    def __init__(self, **kwargs):

        super(ProfileViewerState, self).__init__()

        self.ref_data_helper = ManualDataComboHelper(self, 'reference_data')

        self.add_callback('layers', self._layers_changed)
        self.add_callback('reference_data', self._reference_data_changed, echo_old=True)
        self.add_callback('x_att', self._update_att)
        self.add_callback('normalize', self._reset_y_limits)

        self.x_att_helper = ComponentIDComboHelper(self, 'x_att',
                                                   numeric=False, datetime=False, categorical=False,
                                                   pixel_coord=True)

        ProfileViewerState.function.set_choices(self, list(FUNCTIONS))
        ProfileViewerState.function.set_display_func(self, FUNCTIONS.get)

        self.update_from_dict(kwargs)

    def _update_combo_ref_data(self):
        self.ref_data_helper.set_multiple_data(self.layers_data)

    def reset_limits(self):
        with delay_callback(self, 'x_min', 'x_max', 'y_min', 'y_max'):
            self._reset_x_limits()
            self._reset_y_limits()

    @property
    def _display_world(self):
        return getattr(self.reference_data, 'coords', None) is not None

    @defer_draw
    def _update_att(self, *args):
        if self.x_att is not None:
            if self._display_world:
                if self.x_att in self.reference_data.pixel_component_ids:
                    self.x_att_pixel = self.x_att
                else:
                    index = self.reference_data.world_component_ids.index(self.x_att)
                    self.x_att_pixel = self.reference_data.pixel_component_ids[index]
            else:
                self.x_att_pixel = self.x_att
        self._reset_x_limits()

    def _reset_x_limits(self, *event):

        # NOTE: we don't use AttributeLimitsHelper because we need to avoid
        # trying to get the minimum of *all* the world coordinates in the
        # dataset. Instead, we use the same approach as in the layer state below
        # and in the case of world coordinates we use online the spine of the
        # data.

        if self.reference_data is None or self.x_att_pixel is None:
            return

        data = self.reference_data

        if self.x_att in data.pixel_component_ids:
            x_min, x_max = -0.5, data.shape[self.x_att.axis] - 0.5
        else:
            axis = data.world_component_ids.index(self.x_att)
            axis_view = [0] * data.ndim
            axis_view[axis] = slice(None)
            axis_values = data[self.x_att, tuple(axis_view)]
            x_min, x_max = np.nanmin(axis_values), np.nanmax(axis_values)

        with delay_callback(self, 'x_min', 'x_max'):
            self.x_min = x_min
            self.x_max = x_max

    def _reset_y_limits(self, *event):
        if self.normalize:
            with delay_callback(self, 'y_min', 'y_max'):
                self.y_min = -0.1
                self.y_max = +1.1
        else:
            y_min, y_max = np.inf, -np.inf
            for layer in self.layers:
                try:
                    profile = layer.profile
                except Exception:  # e.g. incompatible subset
                    continue
                if profile is not None:
                    x, y = profile
                    if len(y) > 0:
                        y_min = min(y_min, nanmin(y))
                        y_max = max(y_max, nanmax(y))
            with delay_callback(self, 'y_min', 'y_max'):
                if y_max > y_min:
                    self.y_min = y_min
                    self.y_max = y_max
                else:
                    self.y_min = 0
                    self.y_max = 1

    def flip_x(self):
        """
        Flip the x_min/x_max limits.
        """
        with delay_callback(self, 'x_min', 'x_max'):
            self.x_min, self.x_max = self.x_max, self.x_min

    @defer_draw
    def _layers_changed(self, *args):
        self._update_combo_ref_data()

    @defer_draw
    def _reference_data_changed(self, before=None, after=None):

        # A callback event for reference_data is triggered if the choices change
        # but the actual selection doesn't - so we avoid resetting the WCS in
        # this case.
        if before is after:
            return

        for layer in self.layers:
            layer.reset_cache()

        # This signal can get emitted if just the choices but not the actual
        # reference data change, so we check here that the reference data has
        # actually changed
        if self.reference_data is not getattr(self, '_last_reference_data', None):
            self._last_reference_data = self.reference_data

            with delay_callback(self, 'x_att'):

                if self.reference_data is None:
                    self.x_att_helper.set_multiple_data([])
                else:
                    self.x_att_helper.set_multiple_data([self.reference_data])
                    if self._display_world:
                        self.x_att_helper.world_coord = True
                        self.x_att = self.reference_data.world_component_ids[0]
                    else:
                        self.x_att_helper.world_coord = False
                        self.x_att = self.reference_data.pixel_component_ids[0]

                self._update_att()

        self.reset_limits()

    def _update_priority(self, name):
        if name == 'layers':
            return 2
        elif name == 'reference_data':
            return 1.5
        elif name.endswith(('_min', '_max')):
            return 0
        else:
            return 1


class ProfileLayerState(MatplotlibLayerState):
    """
    A state class that includes all the attributes for layers in a Profile plot.
    """

    linewidth = DDCProperty(1, docstring='The width of the line')

    attribute = DDSCProperty(docstring='The attribute shown in the layer')
    v_min = DDCProperty(docstring='The lower level shown')
    v_max = DDCProperty(docstring='The upper level shown')
    percentile = DDSCProperty(docstring='The percentile value used to '
                                        'automatically calculate levels')

    _viewer_callbacks_set = False
    _profile_cache = None

    def __init__(self, layer=None, viewer_state=None, **kwargs):

        super(ProfileLayerState, self).__init__(layer=layer, viewer_state=viewer_state)

        self.attribute_att_helper = ComponentIDComboHelper(self, 'attribute',
                                                           numeric=True, categorical=False)

        percentile_display = {100: 'Min/Max',
                              99.5: '99.5%',
                              99: '99%',
                              95: '95%',
                              90: '90%',
                              'Custom': 'Custom'}

        ProfileLayerState.percentile.set_choices(self, [100, 99.5, 99, 95, 90, 'Custom'])
        ProfileLayerState.percentile.set_display_func(self, percentile_display.get)

        self.add_callback('layer', self._update_attribute, priority=1000)

        if layer is not None:
            self._update_attribute()

        self.update_from_dict(kwargs)

    def _update_attribute(self, *args):
        if self.layer is not None:
            self.attribute_att_helper.set_multiple_data([self.layer])

    @property
    def independent_x_att(self):
        return is_convertible_to_single_pixel_cid(self.layer, self.viewer_state.x_att) is not None

    def normalize_values(self, values):
        return (np.asarray(values) - self.v_min) / (self.v_max - self.v_min)

    def reset_cache(self, *args):
        self._profile_cache = None

    @property
    def viewer_state(self):
        return self._viewer_state

    @viewer_state.setter
    def viewer_state(self, viewer_state):
        self._viewer_state = viewer_state

    @property
    def profile(self):
        self.update_profile()
        return self._profile_cache

    def update_profile(self, update_limits=True):

        if self._profile_cache is not None:
            return self._profile_cache

        if not self.visible:
            return

        if not self._viewer_callbacks_set:
            self.viewer_state.add_callback('x_att', self.reset_cache, priority=100000)
            self.viewer_state.add_callback('function', self.reset_cache, priority=100000)
            if self.is_callback_property('attribute'):
                self.add_callback('attribute', self.reset_cache, priority=100000)
            self._viewer_callbacks_set = True

        if self.viewer_state is None or self.viewer_state.x_att is None or self.attribute is None:
            raise IncompatibleDataException()

        # Check what pixel axis in the current dataset x_att corresponds to
        pix_cid = is_convertible_to_single_pixel_cid(self.layer, self.viewer_state.x_att_pixel)

        if pix_cid is None:
            raise IncompatibleDataException()

        # If we get here, then x_att does correspond to a single pixel axis in
        # the cube, so we now prepare a list of axes to collapse over.
        axes = tuple(i for i in range(self.layer.ndim) if i != pix_cid.axis)

        # We now get the y values for the data

        # TODO: in future we should optimize the case where the mask is much
        # smaller than the data to just average the relevant 'spaxels' in the
        # data rather than collapsing the whole cube.

        if isinstance(self.layer, Subset):
            data = self.layer.data
            subset_state = self.layer.subset_state
        else:
            data = self.layer
            subset_state = None

        profile_values = data.compute_statistic(self.viewer_state.function, self.attribute, axis=axes, subset_state=subset_state)

        if np.all(np.isnan(profile_values)):
            self._profile_cache = [], []
        else:
            axis_view = [0] * data.ndim
            axis_view[pix_cid.axis] = slice(None)
            axis_values = data[self.viewer_state.x_att, tuple(axis_view)]
            self._profile_cache = axis_values, profile_values

        if update_limits:
            self.update_limits(update_profile=False)

    def update_limits(self, update_profile=True):
        with delay_callback(self, 'v_min', 'v_max'):
            if update_profile:
                self.update_profile(update_limits=False)
            if self._profile_cache is not None and len(self._profile_cache[1]) > 0:
                self.v_min = nanmin(self._profile_cache[1])
                self.v_max = nanmax(self._profile_cache[1])



import time

from qtpy import QtCore
from glue.core import Data
from glue.core.tests.util import simple_session

from ..style_dialog import StyleDialog


class NonBlockingStyleDialog(StyleDialog):
    def exec_(self, *args):
        self.show()
        time.sleep(0.1)
        self.accept()


def test_style_dialog():

    # This is in part a regression test for a bug in Python 3. It is not a
    # full test of StyleDialog.

    session = simple_session()
    hub = session.hub
    collect = session.data_collection

    image = Data(label='im',
                 x=[[1, 2], [3, 4]],
                 y=[[2, 3], [4, 5]])

    pos = QtCore.QPoint(10, 10)
    st = NonBlockingStyleDialog.dropdown_editor(image, pos)



from qtpy import QtCore

__all__ = ['PyMimeData']


class PyMimeData(QtCore.QMimeData):
    """
    A custom MimeData object that stores live python objects

    Associate specific objects with a mime type by passing
    mime type / object key/value pairs to the __init__ method

    If a single object is passed to the init method, that
    object is associated with the PyMimeData.MIME_TYPE mime type
    """
    MIME_TYPE = 'application/py_instance'

    def __init__(self, instance=None, **kwargs):
        """
        :param instance: The python object to store

        kwargs: Optional mime type / objects pairs to store as objects
        """
        super(PyMimeData, self).__init__()

        self._instances = {}

        self.setData(self.MIME_TYPE, instance)
        for k, v in kwargs.items():
            self.setData(k, v)

    def formats(self):
        return list(set(super(PyMimeData, self).formats() +
                        list(self._instances.keys())))

    def hasFormat(self, fmt):
        return fmt in self._instances or super(PyMimeData, self).hasFormat(fmt)

    def setData(self, mime, data):
        super(PyMimeData, self).setData(mime, QtCore.QByteArray(1, '1'))
        self._instances[mime] = data

    def data(self, mime_type):
        """ Retrieve the data stored at the specified mime_type

        If mime_type is application/py_instance, a python object
        is returned. Otherwise, a QtCore.QByteArray is returned """

        if str(mime_type) in self._instances:
            return self._instances[mime_type]

        return super(PyMimeData, self).data(mime_type)



# pylint: disable=I0011,W0613,W0201,W0212,E1101,E1103

import pytest
from unittest.mock import MagicMock

from ..data import Data
from ..data_collection import DataCollection
from ..exceptions import InvalidSubscriber, InvalidMessage
from ..hub import Hub, HubListener
from ..message import SubsetMessage, Message
from ..subset import Subset


class TestHub(object):

    def setup_method(self, method):
        self.hub = Hub()

    def get_subscription(self):
        msg = Message
        handler = MagicMock()
        subscriber = MagicMock(spec_set=HubListener)
        return msg, handler, subscriber

    def test_subscribe(self):
        msg, handler, subscriber = self.get_subscription()
        self.hub.subscribe(subscriber, msg, handler)
        assert self.hub.is_subscribed(subscriber, msg)
        assert self.hub.get_handler(subscriber, msg) == handler

    def test_get_handler(self):
        msg, handler, subscriber = self.get_subscription()
        self.hub.subscribe(subscriber, msg, handler)
        assert self.hub.get_handler(subscriber, msg) == handler
        assert self.hub.get_handler(subscriber, None) is None
        assert self.hub.get_handler(None, msg) is None

    def test_unsubscribe(self):
        msg, handler, subscriber = self.get_subscription()
        self.hub.subscribe(subscriber, msg, handler)
        self.hub.unsubscribe(subscriber, msg)

        assert not self.hub.is_subscribed(subscriber, msg)
        assert self.hub.get_handler(subscriber, msg) is None

    def test_unsubscribe_all(self):
        msg, handler, subscriber = self.get_subscription()
        msg2 = SubsetMessage
        self.hub.subscribe(subscriber, msg, handler)
        self.hub.subscribe(subscriber, msg2, handler)
        self.hub.unsubscribe_all(subscriber)
        assert not self.hub.is_subscribed(subscriber, msg)
        assert not self.hub.is_subscribed(subscriber, msg2)

    def test_unsubscribe_specific_to_message(self):
        msg, handler, subscriber = self.get_subscription()
        msg2 = SubsetMessage
        self.hub.subscribe(subscriber, msg, handler)
        self.hub.subscribe(subscriber, msg2, handler)
        self.hub.unsubscribe(subscriber, msg)
        assert not self.hub.is_subscribed(subscriber, msg)
        assert self.hub.is_subscribed(subscriber, msg2)

    def test_broadcast(self):
        msg, handler, subscriber = self.get_subscription()
        self.hub.subscribe(subscriber, msg, handler)
        msg_instance = msg("Test")
        self.hub.broadcast(msg_instance)
        handler.assert_called_once_with(msg_instance)

    def test_unsubscribe_halts_broadcast(self):
        msg, handler, subscriber = self.get_subscription()
        self.hub.subscribe(subscriber, msg, handler)
        self.hub.unsubscribe(subscriber, msg)
        msg_instance = msg("Test")
        self.hub.broadcast(msg_instance)
        assert handler.call_count == 0

    def test_unsubscribe_spec_setific_to_message(self):
        msg, handler, subscriber = self.get_subscription()
        msg2 = SubsetMessage
        self.hub.subscribe(subscriber, msg2, handler)
        msg_instance = msg("Test")
        self.hub.broadcast(msg_instance)
        assert handler.call_count == 0

    def test_subscription_catches_message_subclasses(self):
        msg, handler, subscriber = self.get_subscription()
        msg2 = SubsetMessage
        self.hub.subscribe(subscriber, msg, handler)
        msg_instance = msg2(MagicMock(spec_set=Subset))
        self.hub.broadcast(msg_instance)
        handler.assert_called_once_with(msg_instance)

    def test_handler_ignored_if_subset_handler_present(self):
        msg, handler, subscriber = self.get_subscription()
        handler2 = MagicMock()
        msg2 = SubsetMessage
        self.hub.subscribe(subscriber, msg, handler)
        self.hub.subscribe(subscriber, msg2, handler2)
        msg_instance = SubsetMessage(Subset(None))
        self.hub.broadcast(msg_instance)
        handler2.assert_called_once_with(msg_instance)
        assert handler.call_count == 0

    def test_filter(self):
        msg, handler, subscriber = self.get_subscription()
        filter = lambda x: False
        self.hub.subscribe(subscriber, msg, handler)
        msg_instance = msg("Test")
        self.hub.broadcast(msg)
        assert handler.call_count == 0

    def test_broadcast_sends_to_all_subsribers(self):
        msg, handler, subscriber = self.get_subscription()
        msg, handler2, subscriber2 = self.get_subscription()

        self.hub.subscribe(subscriber, msg, handler)
        self.hub.subscribe(subscriber2, msg, handler2)
        msg_instance = msg("Test")
        self.hub.broadcast(msg_instance)
        handler.assert_called_once_with(msg_instance)
        handler2.assert_called_once_with(msg_instance)

    def test_invalid_unsubscribe_ignored(self):
        msg, handler, subscriber = self.get_subscription()
        self.hub.unsubscribe(handler, subscriber)

    def test_invalid_subscribe(self):

        msg, handler, subscriber = self.get_subscription()

        with pytest.raises(InvalidSubscriber) as exc:
            self.hub.subscribe(None, msg, handler)
        assert exc.value.args[0].startswith("Subscriber must be a HubListener")

        with pytest.raises(InvalidMessage) as exc:
            self.hub.subscribe(subscriber, None, handler)
        assert exc.value.args[0].startswith("message class must be "
                                            "a subclass of glue.Message")

    def test_default_handler(self):
        msg, handler, subscriber = self.get_subscription()
        self.hub.subscribe(subscriber, msg)
        msg_instance = msg("Test")

        self.hub.broadcast(msg_instance)
        subscriber.notify.assert_called_once_with(msg_instance)

    def test_autosubscribe(self):
        l = MagicMock(spec_set=HubListener)
        d = MagicMock(spec_set=Data)
        s = MagicMock(spec_set=Subset)
        dc = MagicMock(spec_set=DataCollection)
        hub = Hub(l, d, s, dc)

        l.register_to_hub.assert_called_once_with(hub)
        d.register_to_hub.assert_called_once_with(hub)
        dc.register_to_hub.assert_called_once_with(hub)
        s.register.assert_called_once_with()

    def test_invalid_init(self):
        with pytest.raises(TypeError) as exc:
            Hub(None)
        assert exc.value.args[0] == ("Inputs must be HubListener, data, "
                                     "subset, or data collection objects")


class TestHubListener(object):
    """This is a dumb test, I know. Fixated on code coverage"""

    def test_unimplemented(self):
        hl = HubListener()
        with pytest.raises(NotImplementedError):
            hl.register_to_hub(None)
        with pytest.raises(NotImplementedError):
            hl.notify(None)



import os
import weakref
import platform
from collections import OrderedDict

import numpy as np
from matplotlib.colors import ColorConverter

from qtpy import QtWidgets
from glue.core.message import SettingsChangeMessage
from glue.utils.qt import load_ui, ColorProperty, get_qapp
from glue.utils.qt.widget_properties import (CurrentComboTextProperty,
                                             ValueProperty, ButtonProperty)
from glue._settings_helpers import save_settings

__all__ = ["PreferencesDialog"]

rgb = ColorConverter().to_rgb

AUTOLINK_OPTIONS = OrderedDict([
    ('always_show', 'Always show suggestions'),
    ('always_accept', 'Always accept suggestions'),
    ('always_ignore', 'Always ignore suggestions')
])


class AutolinkPreferencesPane(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(AutolinkPreferencesPane, self).__init__(parent=parent)

        from glue.config import settings, autolinker  # noqa

        layout = QtWidgets.QGridLayout()

        self.combos = {}

        if len(autolinker) > 0:

            for i, (label, _) in enumerate(autolinker):
                combo = QtWidgets.QComboBox()
                for short, display in AUTOLINK_OPTIONS.items():
                    combo.addItem(display, userData=short)
                if label in settings.AUTOLINK:
                    index = list(AUTOLINK_OPTIONS.keys()).index(settings.AUTOLINK[label])
                else:
                    index = 0
                combo.setCurrentIndex(index)
                layout.addWidget(QtWidgets.QLabel(label), i, 0)
                layout.addWidget(combo, i, 1)
                self.combos[label] = combo

            layout.addWidget(QtWidgets.QWidget(), i + 1, 0)

        self.setLayout(layout)

    def finalize(self):
        from glue.config import settings
        for label, combo in self.combos.items():
            settings.AUTOLINK[label] = combo.currentData()


class PreferencesDialog(QtWidgets.QDialog):

    theme = CurrentComboTextProperty('ui.combo_theme')
    background = ColorProperty('ui.color_background')
    foreground = ColorProperty('ui.color_foreground')
    data_color = ColorProperty('ui.color_default_data')
    data_alpha = ValueProperty('ui.slider_alpha', value_range=(0, 1))
    data_apply = ButtonProperty('ui.checkbox_apply')
    save_to_disk = ButtonProperty('ui.checkbox_save')
    font_size = ValueProperty('ui.spinner_font_size')

    def __init__(self, application, parent=None):

        super(PreferencesDialog, self).__init__(parent=parent)

        self._app = weakref.ref(application)

        self.ui = load_ui('preferences.ui', self,
                          directory=os.path.dirname(__file__))

        self.ui.cancel.clicked.connect(self.reject)
        self.ui.ok.clicked.connect(self.accept)

        self.ui.combo_theme.currentIndexChanged.connect(self._update_colors_from_theme)

        self.ui.button_reset_dialogs.clicked.connect(self._reset_dialogs)

        # The following is needed because of a bug in Qt which means that
        # tab titles don't get scaled right.
        if platform.system() == 'Darwin':
            app = get_qapp()
            app_font = app.font()
            self.ui.tab_widget.setStyleSheet('font-size: {0}px'.format(app_font.pointSize()))

        from glue.config import settings
        self.background = settings.BACKGROUND_COLOR
        self.foreground = settings.FOREGROUND_COLOR
        self.data_color = settings.DATA_COLOR
        self.data_alpha = settings.DATA_ALPHA
        self.font_size = settings.FONT_SIZE

        self._update_theme_from_colors()

        self._autolink_pane = AutolinkPreferencesPane()
        self.ui.tab_widget.addTab(self._autolink_pane, 'Autolinking')

        self.panes = []

        from glue.config import preference_panes
        for label, widget_cls in sorted(preference_panes):
            pane = widget_cls()
            self.ui.tab_widget.addTab(pane, label)
            self.panes.append(pane)

    def _update_theme_from_colors(self, *args):
        if (rgb(self.background) == (1, 1, 1) and rgb(self.foreground) == (0, 0, 0) and
                rgb(self.data_color) == (0.35, 0.35, 0.35) and np.allclose(self.data_alpha, 0.8)):
            self.theme = 'Black on White'
        elif (rgb(self.background) == (0, 0, 0) and rgb(self.foreground) == (1, 1, 1) and
                rgb(self.data_color) == (0.75, 0.75, 0.75) and np.allclose(self.data_alpha, 0.8)):
            self.theme = 'White on Black'
        else:
            self.theme = 'Custom'

    def _update_colors_from_theme(self, *args):
        if self.theme == 'Black on White':
            self.foreground = 'black'
            self.background = 'white'
            self.data_color = '0.35'
            self.data_alpha = 0.8
        elif self.theme == 'White on Black':
            self.foreground = 'white'
            self.background = 'black'
            self.data_color = '0.75'
            self.data_alpha = 0.8
        elif self.theme != 'Custom':
            raise ValueError("Unknown theme: {0}".format(self.theme))

    def _reset_dialogs(self, *args):
        from glue.config import settings
        for key, _, _ in settings:
            if key.lower().startswith(('show_info', 'show_warn', 'show_large')):
                setattr(settings, key, True)

    def accept(self):

        # Update default settings

        from glue.config import settings
        settings.FOREGROUND_COLOR = self.foreground
        settings.BACKGROUND_COLOR = self.background
        settings.DATA_COLOR = self.data_color
        settings.DATA_ALPHA = self.data_alpha
        settings.FONT_SIZE = self.font_size

        self._autolink_pane.finalize()

        for pane in self.panes:
            pane.finalize()

        # Save to disk if requested
        if self.save_to_disk:
            save_settings()
        else:
            settings._save_to_disk = True

        # Trigger viewers to update defaults

        app = self._app()

        if app is not None:
            app._hub.broadcast(SettingsChangeMessage(self, ('FOREGROUND_COLOR', 'BACKGROUND_COLOR', 'FONT_SIZE')))
            if self.data_apply:  # If requested, trigger data to update color
                app.set_data_color(settings.DATA_COLOR, settings.DATA_ALPHA)

        super(PreferencesDialog, self).accept()


if __name__ == "__main__":

    app = get_qapp()
    widget = PreferencesDialog()
    widget.show()
    widget.raise_()
    app.exec_()
